@@ -10,6 +10,9 @@
    the migration function will be invoked, and the schema version and current
    time will be recorded in the schema_migrations table.
 
+   A migration function can return a map with ::vacuum-analyze to indicate what tables
+   need to be analyzed post-migration.
+
    NOTE: in order to support bug-fix schema changes to older branches without
    breaking the ability to upgrade, it is possible to define a sequence of
    migrations with non-sequential integers.  e.g., if the 1.0.x branch
@@ -59,7 +62,11 @@
             [clj-time.core :refer [now]]
             [puppetlabs.puppetdb.jdbc :as jdbc :refer [query-to-vec]]
             [puppetlabs.puppetdb.config :as conf]
-            [puppetlabs.i18n.core :refer [trs]]))
+            [puppetlabs.i18n.core :refer [trs]]
+            [puppetlabs.puppetdb.scf.hash :as hash]
+            [clojure.set :as set]
+            [clojure.string :as str])
+  (:import [org.postgresql.util PGobject]))
 
 (defn init-through-2-3-8
   []
@@ -1281,6 +1288,7 @@
      (str "ALTER TABLE ONLY edges"
           "  ADD CONSTRAINT edges_certname_source_target_type_unique_key"
           "  UNIQUE (certname, source, target, type)"))))
+
 (defn add-latest-report-timestamp-to-certnames []
   (jdbc/do-commands
     "DROP INDEX IF EXISTS idx_certnames_latest_report_timestamp"
@@ -1301,6 +1309,136 @@
    "create index reports_job_id_idx on reports(job_id) where job_id is not null"
    "create index catalogs_job_id_idx on catalogs(job_id) where job_id is not null"))
 
+(defn maybe-parse-json [s]
+  (some-> s json/parse))
+
+(defn rededuplicate-facts []
+  (log/info (trs "[1/7] Cleaning up unreferenced facts..."))
+  (jdbc/do-commands
+   "DELETE FROM facts WHERE factset_id NOT IN (SELECT id FROM factsets)")
+
+  (log/info (trs "[2/7] Creating new fact storage tables..."))
+  (jdbc/do-commands
+   "CREATE SEQUENCE fact_values_id_seq;"
+
+   ;; will add not-null constraint on value_hash below; we don't have all the
+   ;; values for this yet
+   "CREATE TABLE fact_values (
+      id bigint DEFAULT nextval('fact_values_id_seq'::regclass) NOT NULL,
+      value_hash bytea,
+      value_type_id bigint NOT NULL,
+      value_integer bigint,
+      value_float double precision,
+      value_string text,
+      value_boolean boolean,
+      value jsonb
+    );"
+
+   ;; Do this early to help with the value_hash update queries below
+   "ALTER TABLE fact_values ADD CONSTRAINT fact_values_pkey PRIMARY KEY (id);"
+
+   "CREATE TABLE facts_transform (
+       factset_id bigint NOT NULL,
+       fact_path_id bigint NOT NULL,
+       fact_value_id bigint NOT NULL
+    );")
+
+  (log/info (trs "[3/7] Copying unique fact values into fact_values"))
+  (jdbc/do-commands
+   "INSERT INTO fact_values (value, value_integer, value_float, value_string, value_boolean, value_type_id)
+       SELECT distinct value, value_integer, value_float, value_string, value_boolean, value_type_id FROM facts")
+
+
+  ;; Handle null fv.value separately; allowing them here leads to an intractable
+  ;; query plan
+  (log/info (trs "[4/7] Reconstructing facts to refer to fact_values..."))
+  (jdbc/do-commands
+   "INSERT INTO facts_transform (factset_id, fact_path_id, fact_value_id)
+       SELECT f.factset_id, f.fact_path_id, fv.id
+         FROM facts f
+        INNER JOIN fact_values fv
+                ON fv.value_type_id = f.value_type_id
+               AND fv.value = f.value
+               AND fv.value_integer IS NOT DISTINCT FROM f.value_integer
+               AND fv.value_float   IS NOT DISTINCT FROM f.value_float
+               AND fv.value_string  IS NOT DISTINCT FROM f.value_string
+               AND fv.value_boolean IS NOT DISTINCT FROM f.value_boolean
+        WHERE f.value IS NOT NULL AND fv.value IS NOT NULL"
+
+   "INSERT INTO facts_transform (factset_id, fact_path_id, fact_value_id)
+       SELECT f.factset_id, f.fact_path_id, fv.id
+         FROM facts f
+        INNER JOIN fact_values fv
+                ON fv.value_type_id = f.value_type_id
+               AND fv.value IS NOT DISTINCT FROM f.value
+        WHERE f.value IS NULL AND fv.value IS NULL")
+
+  ;; populate fact_values.value_hash
+  (log/info (trs "[5/7] Computing fact value hashes..."))
+  (jdbc/call-with-query-rows
+   ["select id, value::text from fact_values"]
+   (fn [rows]
+     (doseq [batch (partition-all 500 rows)]
+       (let [ids (map :id batch)
+             hashes (map #(-> (:value %)
+                              maybe-parse-json
+                              hash/generic-identity-hash
+                              sutils/munge-hash-for-storage)
+                         batch)]
+         (jdbc/do-prepared
+          "update fact_values set value_hash = in_data.hash
+            from (select unnest(?) as id, unnest(?) as hash) in_data
+            where fact_values.id = in_data.id"
+          [(sutils/array-to-param "bigint" Long ids)
+           (sutils/array-to-param "bytea" PGobject hashes)])))))
+
+
+  (log/info (trs "[6/7] Indexing fact_values table..."))
+  (jdbc/do-commands
+   "DROP TABLE facts"
+   "ALTER TABLE facts_transform rename to facts"
+
+   "ALTER TABLE fact_values alter column value_hash set not null"
+   "ALTER TABLE fact_values ADD CONSTRAINT fact_values_value_hash_key UNIQUE (value_hash);"
+   "CREATE INDEX fact_values_value_float_idx ON fact_values USING btree (value_float);"
+   "CREATE INDEX fact_values_value_integer_idx ON fact_values USING btree (value_integer);")
+
+  (log/info (trs "[7/7] Indexing facts table..."))
+  (jdbc/do-commands
+   "ALTER TABLE facts ADD CONSTRAINT facts_factset_id_fact_path_id_fact_key UNIQUE (factset_id, fact_path_id);"
+   "CREATE INDEX facts_fact_path_id_idx ON facts USING btree (fact_path_id);"
+   "CREATE INDEX facts_fact_value_id_idx ON facts USING btree (fact_value_id);"
+
+   "ALTER TABLE facts ADD CONSTRAINT fact_path_id_fk
+      FOREIGN KEY (fact_path_id)
+      REFERENCES fact_paths(id);"
+
+   "ALTER TABLE facts ADD CONSTRAINT fact_value_id_fk
+      FOREIGN KEY (fact_value_id)
+      REFERENCES fact_values(id)
+      ON UPDATE RESTRICT
+      ON DELETE RESTRICT;"
+
+   "ALTER TABLE facts ADD CONSTRAINT factset_id_fk
+     FOREIGN KEY (factset_id)
+     REFERENCES factsets(id)
+     ON UPDATE CASCADE
+     ON DELETE CASCADE;")
+
+  {::vacuum-analyze #{"facts" "fact_values" "fact_paths"}})
+
+(defn varchar-columns-to-text []
+  (jdbc/do-commands
+    "alter table reports
+     alter column puppet_version type text,
+     alter column configuration_version type text"
+
+    "alter table resource_events
+     alter column status type text,
+     alter column property type text,
+     alter column containing_class type text,
+     alter column file type text"))
+
 (def migrations
   "The available migrations, as a map from migration version to migration function."
   {28 init-through-2-3-8
@@ -1341,7 +1479,9 @@
    60 fix-missing-edges-fk-constraint
    61 add-latest-report-timestamp-to-certnames
    62 reports-partial-indices
-   63 add-job-id})
+   63 add-job-id
+   64 rededuplicate-facts
+   65 varchar-columns-to-text})
 
 (def desired-schema-version (apply max (keys migrations)))
 
@@ -1447,12 +1587,21 @@
                    unexpected))))
 
     (if-let [pending (seq (pending-migrations))]
-      (do
-        (jdbc/with-db-transaction []
-          (doseq [[version migration] pending]
-            (log/info (trs "Applying database migration version {0}" version))
-            (sql-or-die (fn [] (migration) (record-migration! version)))))
-        (sutils/analyze-small-tables small-tables)
+      (let [tables-to-analyze (jdbc/with-db-transaction []
+                                (->> pending
+                                     (map (fn [[version migration]]
+                                            (log/info (trs "Applying database migration version {0}" version))
+                                            (sql-or-die (fn [] (let [result (migration)]
+                                                                    (record-migration! version)
+                                                                    result)))))
+                                     (filter map?)
+                                     (map ::vacuum-analyze)
+                                     (apply set/union small-tables)
+                                     sort))]
+        (log/info (trs "Updating table statistics for: {0}" (str/join ", " tables-to-analyze)))
+        (->> tables-to-analyze
+             (map #(str "vacuum analyze " %))
+             (apply jdbc/do-commands-outside-txn))
         true)
       (do
         (log/info (trs "There are no pending migrations"))
@@ -1467,13 +1616,11 @@
     (log/info (trs "Creating additional index `fact_paths_path_trgm`"))
     (jdbc/do-commands
      "CREATE INDEX fact_paths_path_trgm ON fact_paths USING gist (path gist_trgm_ops)"))
-  (jdbc/do-commands
-   "drop index if exists fact_values_string_trgm")
-  (when-not (sutils/index-exists? "facts_value_string_trgm")
-    (log/info (trs "Creating additional index `facts_value_string_trgm`"))
+
+  (when-not (sutils/index-exists? "fact_values_string_trgm")
+    (log/info (trs "Creating additional index `fact_values_string_trgm`"))
     (jdbc/do-commands
-     ["create index facts_value_string_trgm on facts"
-      "  using gin (value_string gin_trgm_ops)"]))
+      "CREATE INDEX fact_values_string_trgm ON fact_values USING gin (value_string gin_trgm_ops)"))
 
   (when-not (sutils/index-exists? "packages_name_trgm")
     (log/info (trs "Creating additional index `packages_name_trgm`"))
